from typing import Dict


class BankStatementFinalResultResponse:
    def __init__(self,
                 period: Dict[str, str],
                 account_name: str, account_number: str,
                 total_deposits: float, total_withdrawals: float,
                 opening_balance: float, closing_balance: float,
                 average_monthly_balance: float, excel_file_path: str,
                 salary_excel_file_path: str, selected_bank_name: str,
                 min_salary: float | None, max_salary: float | None,
                 predicted_average_salary: float = 0.00,
                 ):
        self.account_number = account_number
        self.excel_file_path = excel_file_path
        self.salary_excel_file_path = salary_excel_file_path
        self.predicted_average_salary = predicted_average_salary
        self.average_monthly_balance = average_monthly_balance
        self.closing_balance = closing_balance
        self.opening_balance = opening_balance
        self.total_withdrawals = total_withdrawals
        self.total_deposits = total_deposits
        self.account_name = account_name
        self.period = period
        self.selected_bank_name = selected_bank_name
        self.min_salary = min_salary
        self.max_salary = max_salary



class InvalidSalaryRange(Exception):

    def __init__(self):
        self.message = 'The provided salary range is invalid, maximum salary must be greater minimum salary'

    def __str__(self):
        return self.message



import re
from bank_statement_reader.BaseBankStatementReport import BankStatementReport
import pandas as pd


class AccessBankStatement(BankStatementReport):

    def __init__(self, pdf_directory, min_salary, max_salary):
        if pdf_directory is None or pdf_directory == '':
            pdf_directory = "pdfs/access.pdf"
        self.pdf_directory = pdf_directory
        super().__init__(password='', pdf_directory=pdf_directory, min_salary=min_salary, max_salary=max_salary,
                         bank_name='access')

    def get_transactions_table_header_mapping(self):
        return {
            'date': 'Date',
            'transaction_details': 'Transaction Details',
            'reference': 'Reference',
            'value_date': 'Value Date',
            'withdrawals': 'Withdrawals',
            'lodgements': 'Lodgements',
            'balance': 'Balance'
        }

    def get_transactions_table_headers(self, reader):
        return [
            'Date',
            'Transaction Details',
            'Reference',
            'Value Date',
            'Withdrawals',
            'Lodgements',
            'Balance'
        ]

    def pad_header_with_unknown(self, rows, headers):
        if len(rows[0]) > len(headers):
            # Index of "Value Date" in the list
            value_date_index = headers.index('Value Date')
            headers.pop(value_date_index - 1)
            # # Insert the new item before "Value Date"
            # headers.insert(value_date_index, unknown)
            return headers

        else:
            return headers

    def get_transactions_table_rows(self, reader, page=0):
        if page == 0:
            table = reader.pages[page].extract_tables()[1]
            rows_without_header = table[2:]
        else:
            table = reader.pages[page].extract_tables()[0]
            rows_without_header = table[1:]
        modified_rows = [[item.replace('\n', '').strip() if item else '' for item in row] for row in
                         rows_without_header]
        new_trans_rows = []
        pattern = r"\d{1,2}-[A-Z]{3}-\d{4}"
        second_pattern = r"\d{1,2}-[A-Z]{3}-\d{2}"
        for row in modified_rows:
            if page == 0:
                trans_date = row[0]
                value_date = row[4]
                row.pop(3)
            else:
                trans_date = row[0]
                value_date = row[3]
            if re.match(pattern, trans_date) is None:
                continue
            if re.match(second_pattern, value_date) is None:
                continue
            new_trans_rows.append(row)

        return new_trans_rows

    def result(self):
        reader, status, message = self.get_pdf_reader()
        print(message)
        if status == 0 and message is not None:
            raise Exception(message)

        text = self.get_pdf_page_text(reader)
        cleaned_text = self.clean_text(text)

        account_name_extracted = self.get_account_name(cleaned_text)
        statement_period_extracted = self.get_statement_period(cleaned_text)
        account_number_extracted = self.get_account_number(cleaned_text)
        total_withdrawals_extracted = self.get_total_withdrawal(cleaned_text)
        total_deposit_extracted = self.get_total_deposit(cleaned_text)
        opening_balance_extracted = self.get_opening_balance(cleaned_text)
        closing_balance_extracted = self.get_closing_balance(cleaned_text)

        table_headers = self.get_transactions_table_headers(reader)

        num_pages = len(reader.pages)
        trans_rows = []
        for page_num in range(num_pages):
            try:
                # print(page_num)
                new_rows = self.get_transactions_table_rows(reader, page_num)
                if page_num == 0:
                    self.pad_header_with_unknown(new_rows, table_headers)
                trans_rows.extend(new_rows)
            except Exception as e:
                print(page_num)
                print("from result", e)

        if opening_balance_extracted is None:
            opening_balance_extracted = self.convert_to_money(trans_rows[0][6])

        if closing_balance_extracted is None:
            opening_balance_extracted = self.convert_to_money(trans_rows[len(trans_rows) - 1][6])
        formatted_df = self.format_dataframe_columns(table_headers, table_rows=trans_rows)
        average_monthly_balance = self.get_average_monthly_balance(formatted_df)

        return {
            'dataframe': formatted_df,
            'period': statement_period_extracted,
            "account_name": account_name_extracted,
            "account_number": account_number_extracted,
            "total_turn_over_credit": total_deposit_extracted,
            "total_turn_over_debits": total_withdrawals_extracted,
            "opening_balance": opening_balance_extracted,
            "closing_balance": closing_balance_extracted,
            "average_monthly_balance": average_monthly_balance
        }

    def predict_salary_income(self, dataframe, table_headers):
        # Filter the DataFrame to get rows with values within the specified range
        filtered_df = dataframe[(dataframe['Deposits'] >= self.min_salary) & (dataframe['Deposits'] <= self.max_salary)]
        potential_salary = []
        for index, row in filtered_df.iterrows():
            unique = self.is_unique_amount_in_month_year(row, filtered_df)
            if not unique:
                continue
            potential_salary.append([
                row['Transaction Date'],
                row['Description'],
                row['Reference'],
                row['Value Date'],
                row['Withdrawals'],
                row['Deposits'],
                row['Balance'],
            ])
        formatted_salary_df = self.format_dataframe_columns(table_headers, potential_salary)
        return formatted_salary_df

# access_bank_statement_pdf_path = "../pdfs/access.pdf"
#
# bank_statement = AccessBankStatement(access_bank_statement_pdf_path)
#
# result = bank_statement.result()
# print(result)
# exit()



from datetime import datetime
from bank_statement_reader.BaseBankStatementReport import BankStatementReport
import pandas as pd
import re


class UBABankStatement(BankStatementReport):
    def __init__(self, pdf_directory, min_salary, max_salary):
        if pdf_directory is None or pdf_directory == '':
            pdf_directory = "pdfs/uba.pdf"
        super().__init__(password='', pdf_directory=pdf_directory, min_salary=min_salary, max_salary=max_salary,
                         bank_name='uba')

    def predict_salary_income(self, dataframe, table_headers):
        # Filter the DataFrame to get rows with values within the specified range
        filtered_df = dataframe[(dataframe['Deposits'] >= self.min_salary) & (dataframe['Deposits'] <= self.max_salary)]
        potential_salary = []
        # Loop through each unique value and find occurrences
        for index, row in filtered_df.iterrows():
            unique = self.is_unique_amount_in_month_year(row, filtered_df)
            if not unique:
                continue
            potential_salary.append([
                row['Transaction Date'],
                row['Value Date'],
                row['Description'],
                row['Chq No'],
                row['Withdrawals'],
                row['Deposits'],
                row['Balance'],
            ])
        salary_df = self.format_dataframe_columns(table_headers, potential_salary)
        return salary_df

    # matches 'September 28th 2022'
    first_pattern = r'\b(?:January|February|March|April|May|June|July|August|September|October|November' \
                    r'|December)\s+\d{1,2}(?:st|nd|rd|th)?\s+\d{4}\b'
    # matches October 2nd October 1st 2022 2022
    second_pattern = r'\b(?:January|February|March|April|May|June|July|August|September|October|November' \
                     r'|December)\s+\d{1,2}(?:st|nd|rd|th)\s+(' \
                     r'?:January|February|March|April|May|June|July|August|September|October|November' \
                     r'|December)\s+\d{1,2}(?:st|nd|rd|th)\s+\d{4}\s+\d{4}\b'

    def clean_date_v2(self, date_str, _format="%Y-%m-%d"):
        try:
            return pd.to_datetime(date_str, format=_format)
        except ValueError as err:
            print(err)
            print('clean_date_v2:', date_str)
            # return date_str

    def get_account_name(self, text):
        full_name = super().get_account_name(text)
        if full_name is not None:
            return full_name
        pattern = r"(?i)BANK\s+STATEMENT\s+(\w+)\s+(\w+)\s+(\w+)"
        match = re.search(pattern, text, re.IGNORECASE)

        if match:
            first_name = match.group(1)
            middle_name = match.group(2)
            last_name = match.group(3)

            if first_name is not None:
                full_name = first_name

            if middle_name is not None:
                full_name = full_name + ' ' + middle_name

            if last_name is not None:
                full_name = full_name + ' ' + last_name
            return full_name
        else:
            return None

    def get_total_withdrawal(self, text):
        total_withdrawals = super().get_total_withdrawal(text)
        if total_withdrawals is not None:
            return total_withdrawals
        pattern = r'Credits[:.]?\s*(-?[\d,.]+)'
        match = re.search(pattern, text, re.IGNORECASE)
        if match:
            total_withdrawals = match.group(1)
            return self.convert_to_money(total_withdrawals)
        else:
            return None

    def get_total_deposit(self, text):
        total_deposit = super().get_total_deposit(text)
        if total_deposit is not None:
            return total_deposit
        pattern = r'Debits[:.]?\s*(-?[\d,.]+)'
        match = re.search(pattern, text, re.IGNORECASE)
        if match:
            total_deposit = match.group(1)
            return self.convert_to_money(total_deposit)
        else:
            return None

    def get_transactions_table_rows(self, reader, page):
        first_table_format = reader.pages[page].extract_tables()
        second_table_format = reader.pages[page].extract_tables(
            table_settings={"vertical_strategy": "text", "horizontal_strategy": "lines"})
        rows_with_errors = []
        if len(first_table_format) > 0:
            # this works for uba 1 sample file
            tables = first_table_format
            if page == 0:
                table = tables[1]
                rows_without_header = table[1:]
            else:
                table = reader.pages[page].extract_tables()[0]
                rows_without_header = table[1:]
            return rows_without_header
        elif len(second_table_format) > 0:
            # this works for uba2 sample file
            tables = second_table_format
            rows = tables[0]
            new_trans_rows = []
            for row in rows:
                if len(row) < 7:
                    continue
                try:
                    if page == 0:
                        trans_date = row[0].replace('\n', ' ')
                        value_date = row[1].replace('\n', ' ' + row[2] + ' ')
                        row[0] = trans_date
                        row[1] = value_date
                        row.pop(2)
                    else:
                        trans_date = row[0].replace('\n', ' ' + row[1] + ' ')
                        value_date = row[2].replace('\n', ' ' + row[3] + ' ')
                        if re.match(self.first_pattern, trans_date) is None and re.match(self.second_pattern,
                                                                                         trans_date) is None:
                            continue
                        if re.match(self.first_pattern, value_date) is None and re.match(self.second_pattern,
                                                                                         value_date) is None:
                            continue
                        row[0] = trans_date
                        row[2] = value_date
                        row[3] = row[4]  # row 3 is narration
                        row.pop(1)
                        row.pop(3)
                    # Convert the input string to a datetime object

                    row[0] = self.custom_date_format(row[0])
                    row[1] = self.custom_date_format(row[1])
                    if len(row) == len(self.get_transactions_table_header_mapping()):
                        new_trans_rows.append(row)
                except Exception as ex:
                    rows_with_errors.append(row)
                    print("here:", ex)
                    print(row)
                    print("page:", page)
            return new_trans_rows
        else:
            return []

    # formats date in this format  "December 2nd 2022"
    def custom_date_format(self, date_str):
        try:

            if re.match(self.first_pattern, date_str):
                date_str = date_str

            if re.match(self.second_pattern, date_str):
                split_date_str = date_str.replace('\n', ' ').split()
                date_str = split_date_str[0] + ' ' + split_date_str[1] + ' ' + split_date_str[len(split_date_str) - 1]

            # Remove the suffix from the day part
            day_str = date_str.split(' ')[1].rstrip('nd').rstrip('st').rstrip('rd').rstrip('th')
            date_str_without_suffix = f"{date_str.split(' ')[0]} {day_str} {date_str.split(' ')[2]}"

            # Convert the input string to a datetime object
            date_obj = datetime.strptime(date_str_without_suffix, "%B %d %Y")

            # Format the datetime object to your desired output format
            formatted_date = date_obj.strftime("%Y-%m-%d")
            return formatted_date
        except Exception as custom_date_e:
            print("custom_date_format: ", custom_date_e)
            print("date string: ", date_str)
            raise ValueError("Date does not match  any of the patterns")

    def get_transactions_table_header_mapping(self):
        return {
            'trans_date': 'Trans Date',
            'value_date': 'Value Date',
            'narration': 'Narration',
            'chq_no': 'Chq No',
            'debit': 'Debit',  # money withdrawn from account
            'credit': 'Credit',  # money sent to account
            'balance': 'Balance',
        }

    def get_statement_period(self, text):
        parent_period = super().get_statement_period(text)
        if parent_period.get('from_date') is not None and parent_period.get('to_date') is not None:
            return parent_period

        from_date_pattern = r'Transaction Date From[:.]?\s*(\b(?:January|February|March|April|May|June|July|August' \
                            r'|September|October|November|December)\s+\d{1,2}(?:st|nd|rd|th)?\s+\d{4}\b)'

        to_date_pattern = r'Transaction Date To[:.]?\s*(\b(?:January|February|March|April|May|June|July|August' \
                          r'|September|October|November|December)\s+\d{1,2}(?:st|nd|rd|th)?\s+\d{4}\b)'

        match = re.search(from_date_pattern, text, re.IGNORECASE)
        if re.search(from_date_pattern, text, re.IGNORECASE):
            from_date = match.group(1)
            from_date = self.custom_date_format(from_date)
        else:
            from_date = None

        if re.search(to_date_pattern, text, re.IGNORECASE):
            to_date = match.group(1)
            to_date = self.custom_date_format(to_date)
        else:
            to_date = None
        return {'from_date': from_date, 'to_date': to_date}

    def result(self):
        reader, status, message = self.get_pdf_reader()
        print(message)
        if status == 0:
            raise Exception("Reading of file failed")

        text = self.get_pdf_page_text(reader)
        cleaned_text = self.clean_text(text)

        account_name_extracted = self.get_account_name(cleaned_text)
        statement_period_extracted = self.get_statement_period(cleaned_text)
        account_number_extracted = self.get_account_number(cleaned_text)
        total_withdrawals_extracted = self.get_total_withdrawal(cleaned_text)
        total_deposit_extracted = self.get_total_deposit(cleaned_text)
        opening_balance_extracted = self.get_opening_balance(cleaned_text)
        closing_balance_extracted = self.get_closing_balance(cleaned_text)

        table_headers = self.get_transactions_table_headers(reader)

        num_pages = len(reader.pages)
        trans_rows = []
        for page_num in range(num_pages):
            try:
                # print(page_num)
                new_rows = self.get_transactions_table_rows(reader, page_num)
                trans_rows.extend(new_rows)
            except Exception as e:
                print(e)

        if opening_balance_extracted is None:
            opening_balance_extracted = self.convert_to_money(trans_rows[0][6])

        if closing_balance_extracted is None:
            closing_balance_extracted = self.convert_to_money(trans_rows[len(trans_rows) - 1][6])

        formatted_df = self.format_dataframe_columns(table_headers, table_rows=trans_rows)
        average_monthly_balance = self.get_average_monthly_balance(formatted_df.copy())
        return {
            'dataframe': formatted_df,
            'period': statement_period_extracted,
            "account_name": account_name_extracted,
            "account_number": account_number_extracted,
            "total_turn_over_credit": total_deposit_extracted,
            "total_turn_over_debits": total_withdrawals_extracted,
            "opening_balance": opening_balance_extracted,
            "closing_balance": closing_balance_extracted,
            "average_monthly_balance": average_monthly_balance
        }

# if __name__ == "__main__":
#     print("Called")
#
#     pdf_path = "../pdfs/uba3.pdf"
#
#     bank_statement = UBABankStatement(pdf_path)
#
#     result = bank_statement.result()
#
#     print(result)



import re

import pandas as pd

from bank_statement_reader.BaseBankStatementReport import BankStatementReport


class FidelityBankStatement(BankStatementReport):

    def __init__(self, pdf_directory, min_salary, max_salary):
        if pdf_directory is None or pdf_directory == '':
            pdf_directory = "pdfs/fidelity.pdf"
        super().__init__(password='', pdf_directory=pdf_directory, min_salary=min_salary, max_salary=max_salary,
                         bank_name='fidelity')

    def get_account_number(self, text):
        pattern = r'Account:\s+(\d{10,12})'
        match = re.search(pattern, text, re.IGNORECASE)
        if match:
            account_number = match.group(1)
            return account_number
        else:
            return None
        pass

    def get_account_name(self, text):

        pattern = r"(?i)Type:\s+(Savings|Current)\s+(\w+)\s+(\w+)\s+(\w+)"

        match = re.search(pattern, text, re.IGNORECASE)
        full_name = ''
        if match:
            first_name = match.group(2)
            middle_name = match.group(3)
            last_name = match.group(4)

            if first_name is not None:
                full_name = first_name

            if middle_name is not None:
                full_name = full_name + ' ' + middle_name

            if last_name is not None:
                full_name = full_name + ' ' + last_name
            return full_name
        else:
            return None

    def get_total_withdrawal(self, dataframe):
        return dataframe['Withdrawals'].sum()

    def get_total_deposit(self, dataframe):
        return dataframe['Deposits'].sum()

    # def get_opening_balance(self, _formatted_summary_table):
    #     return _formatted_summary_table['opening_balance']

    # def get_closing_balance(self, _formatted_summary_table):
    #     return _formatted_summary_table['closing_balance']

    def get_statement_period(self, text):

        pattern = r"From\s+(\d{1,2}\s+(?:January|February|March|April|May|June|July|August|September|October|November" \
                  r"|December)\s+\d{4})\s+to\s+(\d{1,2}\s+(" \
                  r"?:January|February|March|April|May|June|July|August|September|October|November|December)\s+\d{4})"

        match = re.search(pattern, text, re.IGNORECASE)

        if match:
            from_date = match.group(1)
            to_date = match.group(2)
            from_date = self.try_multiple_date_formats(from_date)
            from_date = from_date.strftime("%Y-%m-%d")
            to_date = self.try_multiple_date_formats(to_date)
            to_date = to_date.strftime("%Y-%m-%d")
            return {'from_date': from_date, 'to_date': to_date}
        else:
            return {'from_date': None, 'to_date': None}

    def get_transactions_table_header_mapping(self):
        return {
            'transaction_date': 'Transaction Date',
            'value_date': 'Value Date',
            'channel': 'Channel',
            'details': 'Details',
            'pay_in': 'Pay In',
            'pay_out': 'Pay Out',
            'balance': 'Balance'
        }

    def get_transactions_table_headers(self, reader):
        return self.get_transactions_table_header_mapping().values()

    def get_transactions_table_rows(self, reader, page=0):
        date_pattern = r'\d{1,2}-([A-Z]|[a-z]){3}-\d{2}'
        # money_pattern = r'^\d{1,3}(?:,\d{3})*(?:\.\d{2})?$'
        money_pattern = r'^\d{1,3}(?:,\d{3})*(?:\.\d{2})$'
        channels = ["others", "instant banking", 'banking', 'instant', "online banking", 'online', "nip transfer"]
        table = reader.pages[page].extract_tables(
            table_settings={"vertical_strategy": "text", "horizontal_strategy": "lines"})[0]
        rows_without_header = table[0:]
        modified_rows = []
        for index, row in enumerate(rows_without_header):
            if len(row) > 0:
                is_date = re.match(date_pattern, row[0])
                if is_date is None:
                    continue

            transaction_date = None
            value_date = None
            credit_or_debit = None
            balance = None
            channel = ''
            details = ''
            if balance is not None:
                continue

            for item_index, item in enumerate(row):
                item = item.replace('\n', ' ')

                if item is not None and re.match(date_pattern, item):
                    if transaction_date is None:
                        transaction_date = item
                    else:
                        if value_date is None:
                            value_date = item
                if item_index == 2 and item.isdigit() and transaction_date is not None:
                    if re.match(date_pattern, row[1] + row[item_index]) and value_date is None:
                        value_date = row[1] + row[item_index]

                elif item is not None and re.match(money_pattern, item):
                    # print(float(item.replace(',', '')).isdecimal(), item.replace(',', ''))
                    if credit_or_debit is None:
                        credit_or_debit = item
                        # print("credit_or_debit", credit_or_debit)
                    else:
                        balance = item
                elif item.lower() in channels and item is not None:
                    channel = item
                elif item is not None and item.strip() != '-' and item != '' and re.match(
                        money_pattern, item) is None and re.match(date_pattern, item) is None and item.replace('\n',
                                                                                                               '').lower() not in channels:
                    # print(item)
                    details = details + item
            if transaction_date is None or value_date is None or credit_or_debit is None or balance is None:
                continue
            modified_rows.append(
                [transaction_date, value_date, channel, details, credit_or_debit,
                 credit_or_debit, balance])
        for index, row in enumerate(modified_rows):
            # row[6] is balance column
            # row[5] is a debit column
            # row[4] is credit column
            current_row = modified_rows[index]
            previous_row = modified_rows[index - 1]
            current_row_balance = current_row[6]
            previous_row_balance = previous_row[6]

            current_row_balance = float(current_row_balance.replace(',', ''))
            previous_row_balance = float(previous_row_balance.replace(',', ''))
            if current_row_balance > previous_row_balance:
                row[5] = 0.00
            else:
                row[4] = 0.00
        return modified_rows

    def result(self):
        reader, status, message = self.get_pdf_reader()
        print(message)
        if status == 0:
            raise Exception("Reading of file failed")

        num_pages = len(reader.pages)
        text = self.get_pdf_page_text(reader)

        cleaned_text = self.clean_text(text)
        last_page_text = self.get_pdf_page_text(reader, num_pages - 1)
        account_name_extracted = self.get_account_name(cleaned_text)
        account_number_extracted = self.get_account_number(cleaned_text)
        statement_period_extracted = self.get_statement_period(cleaned_text)
        opening_balance_extracted = self.get_opening_balance(cleaned_text)
        closing_balance_extracted = self.get_closing_balance(last_page_text)
        table_headers = self.get_transactions_table_headers(reader)
        trans_rows = []

        for page_num in range(num_pages):
            try:
                new_rows = self.get_transactions_table_rows(reader, page_num)
                trans_rows.extend(new_rows)
            except Exception as e:
                print(page_num)
                print("from result", e)

        if opening_balance_extracted is None:
            opening_balance_extracted = trans_rows[0][6]

        if closing_balance_extracted is None:
            opening_balance_extracted = trans_rows[len(trans_rows) - 1][6]
        # print(trans_rows)
        # exit()
        formatted_df = self.format_dataframe_columns(table_headers, table_rows=trans_rows)
        formatted_df_copy = formatted_df.copy()
        total_withdrawals_extracted = self.get_total_withdrawal(formatted_df_copy)
        total_deposit_extracted = self.get_total_deposit(formatted_df_copy)
        average_monthly_balance = self.get_average_monthly_balance(formatted_df_copy)
        return {

            'dataframe': formatted_df,
            'period': statement_period_extracted,
            "account_name": account_name_extracted,
            "account_number": account_number_extracted,
            "total_turn_over_credit": total_deposit_extracted,
            "total_turn_over_debits": total_withdrawals_extracted,
            "opening_balance": opening_balance_extracted,
            "closing_balance": closing_balance_extracted,
            "average_monthly_balance": average_monthly_balance
        }

    def predict_salary_income(self, dataframe, table_headers):
        # Filter the DataFrame to get rows with values within the specified range
        filtered_df = dataframe[(dataframe['Deposits'] >= self.min_salary) & (dataframe['Deposits'] <= self.max_salary)]
        potential_salary = []
        for index, row in filtered_df.iterrows():
            unique = self.is_unique_amount_in_month_year(row, filtered_df)
            if not unique:
                continue
            potential_salary.append([
                row['Transaction Date'],
                row['Value Date'],
                row['Channel'],
                row['Description'],
                row['Deposits'],
                row['Withdrawals'],
                row['Balance'],
            ])
        salary_df = self.format_dataframe_columns(table_headers, potential_salary)
        return salary_df

# if __name__ == "__main__":
#     print("Called")
# access_bank_statement_pdf_path = "../pdfs/fidelity.pdf"
#
# bank_statement = FidelityBankStatement(access_bank_statement_pdf_path)
#
# result = bank_statement.result()
# print(result)
# exit()



from bank_statement_reader.BaseBankStatementReport import BankStatementReport
import re


class FirstBankStatement(BankStatementReport):

    def __init__(self, pdf_directory, password, min_salary, max_salary):
        if pdf_directory is None or pdf_directory == '':
            pdf_directory = "pdfs/firstbank.pdf"
            password = "81054"
        super().__init__(password=password, pdf_directory=pdf_directory, min_salary=min_salary, max_salary=max_salary, bank_name='first')

    def get_opening_balance(self, text):
        pattern = r'Opening Balance[:.]?\s*(-?[\d,.]+)'
        match = re.search(pattern, text, re.IGNORECASE)
        if match:
            opening_balance = match.group(1)
            return opening_balance
        else:
            return None
        pass

    def get_closing_balance(self, text):
        pattern = r'Closing Balance[:.]?\s*(-?[\d,.]+)'
        match = re.search(pattern, text, re.IGNORECASE)
        if match:
            closing_balance = match.group(1)
            return closing_balance
        else:
            return None

    def get_transactions_table_header_mapping(self):
        return {
            'transdate': 'TransDate',
            'reference': 'Reference',
            'transaction_details': 'Transaction Details',
            'valuedate': 'ValueDate',
            'deposit': 'Deposit',
            'withdrawal': 'Withdrawal',
            'balance': 'Balance'
        }

    def get_transactions_table_headers(self, reader):
        return self.get_transactions_table_header_mapping().values()

    def pad_header_with_unknown(self, rows, headers):
        if len(rows[0]) > len(headers):
            # New item to be inserted
            unknown = 'Unknown'

            # Index of "Value Date" in the list
            value_date_index = headers.index('Value Date')

            # Insert the new item before "Value Date"
            headers.insert(value_date_index, unknown)
            return headers
        else:
            return headers

    def get_transactions_table_rows(self, reader, page=0):
        split_balance_bf_balance = 0
        if page == 0:
            table = reader.pages[page].extract_tables()[0]
            rows_without_header = table[1:]
            balance_bf = [item for item in rows_without_header if 'Balance B/F' in item[0]]

            split_balance_bf = balance_bf[0][0].split()
            split_balance_bf_balance = split_balance_bf[len(split_balance_bf) - 1]
            balance_bf[0][2] = 'Balance B/F'
            balance_bf[0][6] = split_balance_bf_balance
        else:
            table = reader.pages[page].extract_tables()[0]
            rows_without_header = table[1:]
        rows_without_header = [item for item in rows_without_header if 'END OF STATEMENT' not in item[0]]
        rows_without_header = [item for item in rows_without_header if 'Balance B/F' not in item[0]]
        for row in rows_without_header:
            split_list = row[0].split()
            last_item_in_split_list = split_list[len(split_list) - 1]
            trans_date = split_list[0]
            reference = ''
            transaction_details = split_list[1] + ' ' + split_list[2] + ' ' + split_list[3] + last_item_in_split_list
            value_date = split_list[4]

            deposit = split_list[5]
            withdrawal = split_list[5]
            balance = split_list[len(split_list) - 2]
            row[0] = trans_date
            row[1] = reference
            row[2] = transaction_details
            row[3] = value_date
            row[4] = deposit
            row[5] = withdrawal
            row[6] = balance
        modified_rows = [[item.replace('\n', '').strip() if item else '' for item in row] for row in
                         rows_without_header]

        for index, row in enumerate(modified_rows):
            current_row = modified_rows[index]
            previous_row = modified_rows[index - 1]
            current_row_balance = current_row[6]
            previous_row_balance = previous_row[6]
            if index != 0:
                if current_row_balance > previous_row_balance:
                    balance = row[4]
                    row[4] = balance
                    row[5] = None
                else:
                    row[4] = None
                    row[5] = balance
            else:
                if current_row_balance > split_balance_bf_balance:
                    balance = row[4]
                    row[4] = balance
                    row[5] = None

        return modified_rows

    def result(self):
        reader, status, message = self.get_pdf_reader()
        print(message)
        if status == 0:
            raise Exception("Reading of file failed")
        text = self.get_pdf_page_text(reader)
        cleaned_text = self.clean_text(text)

        statement_period_extracted = self.get_statement_period(cleaned_text)
        account_name_extracted = self.get_account_name(cleaned_text)
        account_number_extracted = self.get_account_number(cleaned_text)
        total_withdrawals_extracted = self.get_total_withdrawal(cleaned_text)
        total_deposit_extracted = self.get_total_deposit(cleaned_text)
        opening_balance_extracted = self.get_opening_balance(cleaned_text)
        closing_balance_extracted = self.get_closing_balance(cleaned_text)

        table_headers = self.get_transactions_table_headers(reader)

        num_pages = len(reader.pages)
        trans_rows = []
        for page_num in range(num_pages):
            try:
                new_rows = self.get_transactions_table_rows(reader, page_num)
                trans_rows.extend(new_rows)
            except Exception as e:
                print(page_num)
                print("from result", e)
        if opening_balance_extracted is None:
            opening_balance_extracted = trans_rows[0][5]

        if closing_balance_extracted is None:
            closing_balance_extracted = trans_rows[len(trans_rows) - 1][5]
        formatted_df = self.format_dataframe_columns(table_headers, table_rows=trans_rows)
        average_monthly_balance = self.get_average_monthly_balance(formatted_df)

        return {
            'dataframe': formatted_df,
            'period': statement_period_extracted,
            "account_name": account_name_extracted,
            "account_number": account_number_extracted,
            "total_turn_over_credit": total_deposit_extracted,
            "total_turn_over_debits": total_withdrawals_extracted,
            "opening_balance": opening_balance_extracted,
            "closing_balance": closing_balance_extracted,
            "average_monthly_balance": average_monthly_balance,
        }

    def predict_salary_income(self, dataframe, table_headers):
        # Filter the DataFrame to get rows with values within the specified range
        filtered_df = dataframe[(dataframe['Deposits'] >= self.min_salary) & (dataframe['Deposits'] <= self.max_salary)]
        if filtered_df.empty:
            return None

        potential_salary = []
        for index, row in filtered_df.iterrows():
            unique = self.is_unique_amount_in_month_year(row, filtered_df)
            if not unique:
                continue
            potential_salary.append([
                row['Transaction Date'],
                row['Reference'],
                row['Description'],
                row['Value Date'],
                row['Deposits'],
                row['Withdrawals'],
                row['Balance'],
            ])
        salary_df = self.format_dataframe_columns(table_headers, potential_salary)
        return salary_df



import logging

from bank_statement_reader.AccessBankStatement import AccessBankStatement
from bank_statement_reader.FcmbBankStatement import FcmbBankStatement
from bank_statement_reader.FidelityBankStatement import FidelityBankStatement
from bank_statement_reader.FirstBankStatement import FirstBankStatement
from bank_statement_reader.GtBankStatement import GtBankStatement
from bank_statement_reader.SterlingBankStatement import SterlingBankStatement
from bank_statement_reader.UBABankStatement import UBABankStatement
from bank_statement_reader.ZenithBankStatement import ZenithBankStatement
from bank_statement_reader.BankStatementFinalResultResponse import BankStatementFinalResultResponse
from bank_statement_reader.exceptions.InvalidBankStatementChoice import InvalidBankStatementChoice
from bank_statement_reader.exceptions.BankStatementProcessingFailed import BankStatementProcessingFailed
from bank_statement_reader.exceptions.InvalidSalaryRange import InvalidSalaryRange
from bank_statement_reader.BaseBankStatementReport import BankStatementReport


class BankStatementExecutor:
    bank_statement_report_instance: BankStatementReport
    BANK_STATEMENTS_CHOICES = {
        1: "Zenith",
        2: "UBA",
        3: "Access",
        4: "First",
        5: "GT",
        6: "FCMB",
        7: "Fidelity",
        8: "Sterling"
        # Add more bank statements with corresponding numbers here
    }

    def __init__(self):
        # self.pdf_directory = pdf_directory
        self.bank_statements = self.BANK_STATEMENTS_CHOICES

    def close(self):
        exit(0)

    def display_menu(self):
        print("Available Bank Statements:")
        for number, statement in self.bank_statements.items():
            print(f"{number}. {statement}")

    def get_user_choice(self):
        while True:
            try:
                choice = int(input("Select a number to execute the corresponding bank statement: "))
                if choice in self.bank_statements:
                    return choice
                else:
                    print("Invalid option. Please choose a valid number.")
            except ValueError:
                self.display_menu()
                print("Invalid input. Please enter a number.")

    def execute(self, choice, pdf_file='', password='', min_salary=10000,
                max_salary=500000) -> BankStatementFinalResultResponse:
        try:
            file_to_execute = self.bank_statements.get(choice)
            if file_to_execute is None:
                raise InvalidBankStatementChoice
            print(file_to_execute + " Bank Statement Selected")
            bank_statement = None
            if choice == 1:
                bank_statement = ZenithBankStatement(pdf_directory=pdf_file, min_salary=min_salary,
                                                     max_salary=max_salary)
            elif choice == 2:
                bank_statement = UBABankStatement(pdf_directory=pdf_file, min_salary=min_salary,
                                                  max_salary=max_salary)
            elif choice == 3:
                bank_statement = AccessBankStatement(pdf_directory=pdf_file, min_salary=min_salary,
                                                     max_salary=max_salary)
            elif choice == 4:
                bank_statement = FirstBankStatement(pdf_directory=pdf_file, min_salary=min_salary,
                                                    max_salary=max_salary, password=password)
            elif choice == 5:
                bank_statement = GtBankStatement(pdf_directory=pdf_file, min_salary=min_salary,
                                                 max_salary=max_salary)
            elif choice == 6:
                bank_statement = FcmbBankStatement(pdf_directory=pdf_file, min_salary=min_salary,
                                                   max_salary=max_salary)
            elif choice == 7:
                bank_statement = FidelityBankStatement(pdf_directory=pdf_file, min_salary=min_salary,
                                                       max_salary=max_salary)
            elif choice == 8:
                bank_statement = SterlingBankStatement(pdf_directory=pdf_file, min_salary=min_salary,
                                                       max_salary=max_salary)
            self.bank_statement_report_instance = bank_statement
            result = bank_statement.result()
            reader, status, message = bank_statement.get_pdf_reader()
            table_headers = bank_statement.get_transactions_table_headers(reader)
            salary_df = bank_statement.predict_salary_income(result.get('dataframe'), table_headers)
            predicted_salary_average = bank_statement.get_predicted_salary_average(salary_df)

            # bank_statement.predict_repayment_capability(result.get('dataframe'), required_payment=15000.00)

            excel_file_path = bank_statement.export_to_excel(
                dataframe=result.get('dataframe'),
                name=result.get('account_name'),
                start_date=result.get('period').get('from_date'),
                end_date=result.get('period').get('to_date')
            )
            salary_excel_file_path = bank_statement.export_to_excel(
                dataframe=salary_df,
                name=result.get('account_name'),
                start_date=result.get('period').get('from_date'),
                end_date=result.get('period').get('to_date'),
                is_salary=True
            )

            result.pop('dataframe')
            if excel_file_path:
                result.update({'excel_file_path': excel_file_path})
            if salary_excel_file_path:
                result.update({'salary_excel_file_path': salary_excel_file_path})

            response = BankStatementFinalResultResponse(
                selected_bank_name=bank_statement.bank_name,
                min_salary=bank_statement.min_salary,
                max_salary=bank_statement.max_salary,
                period=result.get('period'),
                account_name=result.get('account_name'),
                account_number=result.get('account_number'),
                total_deposits=result.get('total_turn_over_credit'),
                total_withdrawals=result.get('total_turn_over_debits'),
                opening_balance=result.get('opening_balance'),
                closing_balance=result.get('closing_balance'),
                average_monthly_balance=result.get('average_monthly_balance'),
                excel_file_path=result.get('excel_file_path'),
                salary_excel_file_path=result.get('salary_excel_file_path'),
                predicted_average_salary=predicted_salary_average
            )
            return response
        except InvalidBankStatementChoice as e:
            raise e
        except InvalidSalaryRange as e:
            raise e
        except BankStatementProcessingFailed as e:
            raise e
        except BaseException as e:
            logging.error(e)
            raise BankStatementProcessingFailed



import re

import pandas as pd

from bank_statement_reader.BaseBankStatementReport import BankStatementReport


class FcmbBankStatement(BankStatementReport):

    def __init__(self, pdf_directory, min_salary, max_salary):
        if pdf_directory is None or pdf_directory == '':
            pdf_directory = "pdfs/fcmb.pdf"
        super().__init__(password='', pdf_directory=pdf_directory, min_salary=min_salary, max_salary=max_salary,
                         bank_name='fcmb')

    def format_account_summary_table(self, reader, table_index=0, page=0):
        # get first table in first page
        summary_tables = reader.pages[page].extract_tables()[table_index]
        table_dictionary = {}
        for item in summary_tables:
            if len(item) >= 2:
                # Convert the key by replacing spaces with underscores and making it lowercase
                key1 = item[0].replace(' ', '_').replace('.', '').replace(':', '').replace('\n', '_').lower()
                # Set the value as the second item in the list
                value1 = item[1].replace('\n', '_')
                table_dictionary[key1] = value1
            if len(item) >= 4:
                # Convert the key by replacing spaces with underscores and making it lowercase
                key2 = item[2].replace(' ', '_').replace('.', '').replace(':', '').replace('\n', '_').lower()
                # Set the value as the second item in the list
                value2 = item[3].replace('\n', '_')
                table_dictionary[key2] = value2

        return table_dictionary

    def get_account_number(self, _formatted_summary_table):
        return _formatted_summary_table['account_number']

    def get_account_name(self, _formatted_summary_table):
        return _formatted_summary_table['customer_name']

    def get_opening_balance(self, _formatted_summary_table):
        return self.convert_to_money(_formatted_summary_table['opening_balance'])

    def get_closing_balance(self, _formatted_summary_table):
        return self.convert_to_money(_formatted_summary_table['closing_balance'])

    def get_statement_period(self, _formatted_summary_table):
        from_date = self.try_multiple_date_formats(_formatted_summary_table['start_date'])
        to_date = self.try_multiple_date_formats(_formatted_summary_table['end_date'])
        from_date = from_date.strftime("%Y-%m-%d")
        to_date = to_date.strftime("%Y-%m-%d")
        return {'from_date': from_date, 'to_date': to_date}

    def get_transactions_table_header_mapping(self):
        return {
            'txn_date': 'Txn Date',
            'val_date': 'Val Date',
            'remarks': 'Remarks',
            'debit': 'Debit',
            'credit': 'Credit',
            'balance': 'Balance'
        }

    def get_transactions_table_rows(self, reader, page=0):
        date_pattern = r'\d{1,2}-([A-Z]|[a-z]){3}-\d{4}'
        money_pattern = r'\b\d{1,3}(?:,\d{3})*(?:\.\d{2})?\b'
        if page == 0:
            table = reader.pages[page].extract_tables()[4]
            rows_without_header = table[1:]
        else:
            table = reader.pages[page].extract_tables()[0]
            rows_without_header = table[0:]
        modified_rows = []
        # Remove None
        for index, row in enumerate(rows_without_header):
            transaction_date = None
            value_date = None
            credit_or_debit = None
            balance = None
            remark = None

            if balance is not None:
                continue
            for item in row:
                if item is not None and re.match(date_pattern, item):
                    if transaction_date is None:
                        transaction_date = item
                    else:
                        value_date = item
                elif item is not None and re.match(money_pattern, item):
                    if credit_or_debit is None:
                        credit_or_debit = item
                    else:
                        balance = item
                elif item is not None and item.strip() != '-' and item != '' and len(item) >= 5 and re.match(
                        money_pattern, item) is None and re.match(date_pattern, item) is None:
                    remark = item
            if transaction_date is None or value_date is None or credit_or_debit is None or balance is None:
                continue
            modified_rows.append([transaction_date, value_date, remark, credit_or_debit, credit_or_debit, balance])

        for index, row in enumerate(modified_rows):
            # row[5] is balance column
            # row[4] is credit column
            # row[3] is a debit column
            current_row = modified_rows[index]
            previous_row = modified_rows[index - 1]
            current_row_balance = current_row[5]
            previous_row_balance = previous_row[5]

            if current_row[2].find('Opening Balance') != -1 and page == 0:
                current_row_balance = row[3]
                previous_row_balance = row[3]

            current_row_balance = float(current_row_balance.replace(',', ''))
            previous_row_balance = float(previous_row_balance.replace(',', ''))
            if current_row_balance > previous_row_balance:
                row[3] = 0.00
            else:
                row[4] = 0.00
        return modified_rows

    def result(self):
        reader, status, message = self.get_pdf_reader()
        print(message)
        if status == 0:
            raise Exception("Reading of file failed")

        num_pages = len(reader.pages)
        text = self.get_pdf_page_text(reader)
        last_page_text = self.get_pdf_page_text(reader, num_pages - 1)

        formatted_summary_table = self.format_account_summary_table(reader)
        total_withdrawals_extracted = self.get_total_withdrawal(last_page_text)
        total_deposit_extracted = self.get_total_deposit(last_page_text)
        account_name_extracted = self.get_account_name(formatted_summary_table)
        statement_period_extracted = self.get_statement_period(formatted_summary_table)
        account_number_extracted = self.get_account_number(formatted_summary_table)
        opening_balance_extracted = self.get_opening_balance(formatted_summary_table)
        closing_balance_extracted = self.get_closing_balance(formatted_summary_table)

        table_headers = self.get_transactions_table_headers(reader)

        trans_rows = []

        for page_num in range(num_pages):
            try:
                # print(page_num)
                new_rows = self.get_transactions_table_rows(reader, page_num)
                trans_rows.extend(new_rows)
            except Exception as e:
                print(page_num)
                print("from result", e)
        if opening_balance_extracted is None:
            opening_balance_extracted = trans_rows[0][5]

        if closing_balance_extracted is None:
            opening_balance_extracted = trans_rows[len(trans_rows) - 1][5]

        formatted_df = self.format_dataframe_columns(table_headers, table_rows=trans_rows)
        average_monthly_balance = self.get_average_monthly_balance(formatted_df)
        return {
            'dataframe': formatted_df,
            'period': statement_period_extracted,
            "account_name": account_name_extracted,
            "account_number": account_number_extracted,
            "total_turn_over_credit": total_deposit_extracted,
            "total_turn_over_debits": total_withdrawals_extracted,
            "opening_balance": opening_balance_extracted,
            "closing_balance": closing_balance_extracted,
            "average_monthly_balance": average_monthly_balance
        }

    def predict_salary_income(self, dataframe, table_headers):
        # Filter the DataFrame to get rows with values within the specified range
        filtered_df = dataframe[(dataframe['Deposits'] >= self.min_salary) & (dataframe['Deposits'] <= self.max_salary)]

        potential_salary = []
        for index, row in filtered_df.iterrows():
            unique = self.is_unique_amount_in_month_year(row, filtered_df)
            if not unique:
                continue
            potential_salary.append([
                row['Transaction Date'],
                row['Value Date'],
                row['Description'],
                row['Withdrawals'],
                row['Deposits'],
                row['Balance'],
            ])
        salary_df = self.format_dataframe_columns(table_headers, potential_salary)
        return salary_df

# if __name__ == "__main__":
#     print("Called")
#     access_bank_statement_pdf_path = "../pdfs/fcmb.pdf"
#
#     bank_statement = FcmbBankStatement(access_bank_statement_pdf_path)
#
#     result = bank_statement.result()
#     print(result)
#     exit()



from bank_statement_reader.BaseBankStatementReport import BankStatementReport
import pandas as pd
import re


class GtBankStatement(BankStatementReport):

    def __init__(self, pdf_directory, min_salary, max_salary):
        if pdf_directory is None or pdf_directory == '':
            pdf_directory = "pdfs/gt.pdf"
        super().__init__(password='', pdf_directory=pdf_directory, min_salary=min_salary, max_salary=max_salary,
                         bank_name='gt')

    def get_account_number(self, _formatted_summary_table):
        return _formatted_summary_table['account_no']

    def get_total_withdrawal(self, _formatted_summary_table) -> float:
        return self.convert_to_money(_formatted_summary_table['total_debit'])

    def get_total_deposit(self, _formatted_summary_table) -> float:
        return self.convert_to_money(_formatted_summary_table['total_credit'])

    def get_opening_balance(self, _formatted_summary_table):
        return self.convert_to_money(_formatted_summary_table['opening_balance'])

    def get_closing_balance(self, _formatted_summary_table) -> float:
        return self.convert_to_money(_formatted_summary_table['closing_balance'])

    def get_account_name(self, text):

        pattern = r"(?i)CUSTOMER\s+STATEMENT\s+(\w+)\s+(\w+)\s+(\w+)"

        match = re.search(pattern, text, re.IGNORECASE)
        full_name = ''
        if match:
            first_name = match.group(1)
            middle_name = match.group(2)
            last_name = match.group(3)

            if first_name is not None:
                full_name = first_name

            if middle_name is not None:
                full_name = full_name + ' ' + middle_name

            if last_name is not None:
                full_name = full_name + ' ' + last_name
            return full_name
        else:
            return None

    def get_transactions_table_headers(self, reader):

        table = reader.pages[0].extract_tables()[1]
        headers = table[0]
        header_mapping = self.get_transactions_table_header_mapping()
        header_columns = [header_mapping[col.replace('\n', ' ').lower().replace(' ', '_').replace('.', '').strip()]
                          for col in headers if col is not None]
        return header_columns

    def get_transactions_table_rows(self, reader, page):
        date_pattern = r'\d{1,2}-([A-Z]|[a-z]){3}-\d{4}'
        if page == 0:
            table = reader.pages[page].extract_tables()[1]
            rows_without_header = table[1:]
        else:
            table = reader.pages[page].extract_tables()[0]
            rows_without_header = table[0:]
        trans_rows = []
        for row in rows_without_header:
            trans_date = row[0]
            # value_date = row[1]
            if re.match(date_pattern, trans_date) is None:
                continue
            trans_rows.append(row)
        return trans_rows

    def get_transactions_table_header_mapping(self):
        return {
            'trans_date': 'Trans Date',
            'remarks': 'Remarks',
            'reference': 'Reference',
            'value_date': 'Value Date',
            'debits': 'Debits',
            'credits': 'Credits',
            'balance': 'Balance',
            'originating_branch': 'Originating Branch',
        }

    def predict_salary_income(self, dataframe, table_headers):
        # Filter the DataFrame to get rows with values within the specified range
        filtered_df = dataframe[(dataframe['Deposits'] >= self.min_salary) & (dataframe['Deposits'] <= self.max_salary)]
        potential_salary = []
        for index, row in filtered_df.iterrows():
            unique = self.is_unique_amount_in_month_year(row, filtered_df)
            if not unique:
                continue
            potential_salary.append([
                row['Transaction Date'],
                row['Value Date'],
                row['Reference'],
                row['Withdrawals'],
                row['Deposits'],
                row['Balance'],
                row['Originating Branch'],
                row['Description'],
            ])
        formatted_salary_df = self.format_dataframe_columns(table_headers, potential_salary)
        return formatted_salary_df

    def result(self):
        reader, status, message = self.get_pdf_reader()
        print(message)
        if status == 0:
            raise Exception("Reading of file failed")

        text = self.get_pdf_page_text(reader)
        cleaned_text = self.clean_text(text)
        formatted_summary_table = self.format_account_summary_table(reader)
        statement_period_extracted = self.get_statement_period(cleaned_text)
        account_name_extracted = self.get_account_name(cleaned_text)
        account_number_extracted = self.get_account_number(formatted_summary_table)
        total_withdrawals_extracted = self.get_total_withdrawal(formatted_summary_table)
        total_deposit_extracted = self.get_total_deposit(formatted_summary_table)
        opening_balance_extracted = self.get_opening_balance(formatted_summary_table)
        closing_balance_extracted = self.get_closing_balance(formatted_summary_table)

        table_headers = self.get_transactions_table_headers(reader)

        num_pages = len(reader.pages)
        trans_rows = []
        for page_num in range(num_pages):
            try:
                new_rows = self.get_transactions_table_rows(reader, page_num)
                trans_rows.extend(new_rows)
            except Exception as e:
                print(page_num)
                print("from result", e)
        if opening_balance_extracted is None:
            opening_balance_extracted = trans_rows[0][5]

        if closing_balance_extracted is None:
            closing_balance_extracted = trans_rows[len(trans_rows) - 1][5]
        formatted_df = self.format_dataframe_columns(table_headers, table_rows=trans_rows)
        average_monthly_balance = self.get_average_monthly_balance(formatted_df.copy())
        return {
            'dataframe': formatted_df,
            'period': statement_period_extracted,
            "account_name": account_name_extracted,
            "account_number": account_number_extracted,
            "total_turn_over_credit": total_deposit_extracted,
            "total_turn_over_debits": total_withdrawals_extracted,
            "opening_balance": opening_balance_extracted,
            "closing_balance": closing_balance_extracted,
            "average_monthly_balance": average_monthly_balance
        }



from abc import abstractmethod

import pdfminer
import pdfplumber as pdf_reader
import re
from datetime import datetime
import pandas as pd
from pathlib import Path
from bank_statement_reader.exceptions.InvalidSalaryRange import InvalidSalaryRange


class BankStatementReport:
    pdf_directory = ''
    password = None,
    min_salary = 0.00
    max_salary = 0.00
    TRANSACTION_DATE_NAME = 'Transaction Date'
    VALUE_DATE_COLUMN_NAME = 'Value Date'
    DEPOSIT_COLUMN_NAME = 'Deposits'
    WITHDRAWAL_COLUMN_NAME = 'Withdrawals'
    DESCRIPTION_COLUMN_NAME = 'Description'

    date_formats = [
        "%Y-%m-%d",  # "2022-11-28"
        "%d/%m/%Y",  # "11/01/2023"
        "%d-%b-%y",  # "01-FEB-23"
        "%d-%b-%Y",
        "%d-%B-%Y",  # 01-March-2023
        "%B %d %Y",
        "%Y-%m-%dT%H:%M:%S",  # ISO format with time
        "%Y-%m-%d",  # ISO format without time
        "%m/%d/%Y",  # Month/Day/Year
        "%m/%d/%y",  # Month/Day/Year (short year)
        "%d-%m-%Y",  # Day-Month-Year
        "%Y%m%d",  # Basic ISO format without separators
        "%d/%m/%Y",  # Day/Month/Year
        "%d/%m/%y",  # Day/Month/Year (short year)
        "%b %d, %Y",  # Month abbreviation Day, Year (e.g., Jan 01, 2023)
        "%B %d, %Y",  # Month full name Day, Year (e.g., January 01, 2023)
        "%d %b %Y",  # Day Month abbreviation Year (e.g., 01 Jan 2023)
        "%d %B %Y",  # Day Month full name Year (e.g., 01 January 2023)
        "%Y-%m-%dT%H:%M:%S.%f",  # ISO format with microseconds
        "%Y-%m-%dT%H:%M:%S.%fZ",  # ISO format with microseconds and Zulu timezone
        "%Y/%m/%d",  # Year/Month/Day
        "%Y.%m.%d",  # Year.Month.Day
        "%d.%m.%Y",  # Day.Month.Year
        "%Y.%m.%d %H:%M:%S",  # Year.Month.Day Hour:Minute:Second
        "%d.%m.%Y %H:%M:%S",  # Day.Month.Year Hour:Minute:Second
    ]

    def __init__(self, pdf_directory, bank_name="base_bank", min_salary=10000, max_salary=500000, password=None):
        if max_salary < min_salary:
            raise InvalidSalaryRange
        self.pdf_directory = pdf_directory
        self.min_salary = min_salary
        self.max_salary = max_salary
        self.password = password
        self.bank_name = bank_name

    def clean_date(self, column_name, dataframe, _format='%d-%b-%Y'):
        return pd.to_datetime(dataframe[column_name], format=_format, errors='coerce')

    def try_multiple_date_formats(self, date_string) -> datetime | None:
        if date_string is None or date_string == '':
            return date_string
        if isinstance(date_string, datetime):
            return date_string
        for date_format in self.date_formats:
            try:
                parsed_date = datetime.strptime(date_string.strip(), date_format)
                return parsed_date
            except ValueError:
                continue
        print("date_string: ", date_string)
        # If no format matches, raise an exception or return None as needed.
        raise ValueError("Failed to parse the date using any of the given formats.")

    def clean_money(self, column_name, dataframe):
        return pd.to_numeric(dataframe[column_name].str.replace(',', ''), errors='coerce')

    def clean_money_v2(self, amount):
        if isinstance(amount, str):
            return amount.replace(',', '')
        return amount

    def convert_to_money(self, money: str) -> float:
        return float(money.replace(',', ''))

    def rename_column(self, data_frame, formal, new):
        data_frame.rename(columns={formal: new}, inplace=True)

    def clean_column(self, data_frame, column, to_replace='\n', value=' '):
        return data_frame[column].replace(to_replace, value, regex=True)

    def clean_text(self, text) -> str:
        # Remove \n and \t characters using regular expressions

        cleaned_text = re.sub(r'\n', ' ', text)
        cleaned_text = re.sub(r'\t', ' ', cleaned_text)

        # Remove extra whitespaces
        cleaned_text = re.sub(r'\s+', ' ', cleaned_text)

        return cleaned_text

    def get_pdf_reader(self):
        status = 0
        reader = None
        try:
            reader = pdf_reader.open(self.pdf_directory, password=self.password)
            message = 'File loaded successfully!!!!!!'
            status = 1
        except pdfminer.pdfdocument.PDFPasswordIncorrect:
            message = "Please provide a valid password"
        except FileNotFoundError:
            message = "File could not be found"
        except FileExistsError:
            message = "File does not exist"
        return reader, status, message

    def get_statement_period(self, text):
        # matches 01/10/2022 TO 18/01/2023 on zenith sample
        first_pattern = r"(\b\d{2}/\d{2}/\d{4}\b)\s+TO\s+(\b\d{2}/\d{2}/\d{4}\b)"
        # matches "01-Apr-2023 to 30-Jun-2023" on uba, access sample
        second_pattern = r"(\b\d{2}-\w{3}-\d{4}\b)\s+to\s+(\b\d{2}-\w{3}-\d{4}\b)"
        third_pattern = r"(\b\d{2}-\w{3,}-\d{4}\b)\s+to\s+(\b\d{2}-\w{3,}-\d{4}\b)"

        patterns = [first_pattern, second_pattern, third_pattern]
        for pattern in patterns:
            match = re.search(pattern, text, re.IGNORECASE)
            if match:
                from_date = match.group(1)
                to_date = match.group(2)
                try:
                    output_format = "%Y-%m-%d"
                    from_date = self.try_multiple_date_formats(from_date)
                    from_date = from_date.strftime(output_format)
                    to_date = self.try_multiple_date_formats(to_date)
                    to_date = to_date.strftime(output_format)
                    return {'from_date': from_date, 'to_date': to_date}
                except Exception as e:
                    print("Statement Period Error", e)
                finally:
                    return {'from_date': from_date, 'to_date': to_date}
            else:
                continue

        return {'from_date': None, 'to_date': None}

    def format_account_summary_table(self, reader, table_index=0, page=0):
        # get first table in first page
        table = reader.pages[page].extract_tables()[table_index]

        table_dictionary = {}
        for item in table:
            if len(item) >= 2:
                # Convert the key by replacing spaces with underscores and making it lowercase
                key = item[0].replace(' ', '_').replace('.', '').replace(':', '').replace('\n', '_').lower()
                # Set the value as the second item in the list
                value = item[1]
                table_dictionary[key] = value
        return table_dictionary

    def get_pdf_page_text(self, reader, page=0):
        text = reader.pages[page].extract_text()
        # self.write_to_text_file(self.bank_name, text)
        return text;

    def get_pdf_page_tables(self, reader, page=0):
        return reader.pages[page].extract_tables()

    def get_account_name(self, text=None) -> str | None:
        pattern = r"(?i)Account\s+Name\s*[:.]?\s*([a-zA-Z\s]+)"
        match = re.search(pattern, text, re.IGNORECASE)

        if match:
            account_name = match.group(1)
            if account_name is not None:
                # Split the string by spaces
                words = account_name.split()
                # Take the first 3 words and join them back into a single string
                account_name = ' '.join(words[:3])
            return account_name
        else:
            return None

    def get_account_number(self, text) -> str:
        pattern = r'Account\s+(?:No(?:\.|:)?|Number(?:\.|:)?)\s+(\d{10,12})'
        match = re.search(pattern, text, re.IGNORECASE)
        if match:
            account_number = match.group(1)
            return account_number
        else:
            return None
        pass

    def get_opening_balance(self, text):
        pattern = r'Opening Balance[:.]?\s*(-?[\d,.]+)'
        match = re.search(pattern, text, re.IGNORECASE)
        if match:
            opening_balance = match.group(1)
            return self.convert_to_money(opening_balance)
        else:
            return None

    def get_closing_balance(self, text) -> float | None:
        pattern = r'Closing Balance[:.]?\s*(-?[\d,.]+)'
        match = re.search(pattern, text, re.IGNORECASE)
        if match:
            closing_balance = match.group(1)
            return self.convert_to_money(closing_balance)
        else:
            return None

    def get_transactions_table_headers(self, reader):
        return list(self.get_transactions_table_header_mapping().values())

    @abstractmethod
    def get_transactions_table_rows(self, reader, page):
        pass

    @abstractmethod
    def get_transactions_table_header_mapping(self):
        pass

    def get_total_deposit(self, text) -> float:
        pattern = r'Total (?:Deposits|Credits|Credit|Lodgements)(?::)?\s*([\d,.]+)'

        match = re.search(pattern, text, re.IGNORECASE)

        if match:
            total_deposit = match.group(1)
            return self.convert_to_money(total_deposit)
        else:
            return None

    def get_total_withdrawal(self, text) -> float:
        pattern = r'Total (?:Withdrawals|Debit|Debits)(?::)?\s*([\d,.]+)'
        match = re.search(pattern, text, re.IGNORECASE)

        if match:
            total_withdrawals = match.group(1)
            return self.convert_to_money(total_withdrawals)
        else:
            return None

    def validate_date(self, date_text) -> bool:
        try:
            if date_text != datetime.strptime(date_text, "%Y-%m-%d").strftime('%Y-%m-%d'):
                raise ValueError
            return True
        except ValueError:
            return False

    def get_average_monthly_balance(self, dataframe) -> float:
        return dataframe.groupby(
            [dataframe['Transaction Date'].dt.year.rename('Year'),
             dataframe['Transaction Date'].dt.month.rename('Month')], as_index=False
        )['Balance'].mean()

    def categorize_salary(self, dataframe, column_to_categorize):
        # salary_keywords = ['nft']
        salary_keywords = ['pension', 'sal', 'salary', 'fgn ippis', 'fgnipis', 'fgn sal', 'fgn salary', 'nft', 'cpc',
                           'fbn',
                           'subeb']
        dataframe['IsSalary'] = dataframe[column_to_categorize].str.lower().apply(
            lambda x: any(keyword in x for keyword in salary_keywords))
        salary_transactions = dataframe[dataframe['IsSalary']]
        return salary_transactions

    def categorize_by_credit_amount(self, dataframe, lower_bound, upper_bound, column_to_categorize):

        # Filter the DataFrame to get rows with 'Credit' values within the specified range
        filtered_df = dataframe[
            (dataframe[column_to_categorize] >= lower_bound) & (dataframe[column_to_categorize] <= upper_bound)]

        # Get the unique 'Credit' values within the specified range
        unique_credit_values = filtered_df[column_to_categorize].unique()

        # Find occurrences of each unique credit amount within the range and store in a dictionary
        occurrence_count_dict = {amount: filtered_df[column_to_categorize].eq(amount).sum() for amount in
                                 unique_credit_values}

        # Print the occurrence count of each unique credit amount within the range
        for amount, occurrence_count in occurrence_count_dict.items():
            print(
                f"The occurrence count of {amount} within the range {lower_bound} to {upper_bound} is: {occurrence_count}")

    def format_dataframe_columns(self, table_headers, table_rows):
        if len(table_headers) > 0 and len(table_rows) > 0:
            df = pd.DataFrame(table_rows, columns=table_headers)
            if 'Debits' in table_headers:
                self.rename_column(df, 'Debits', self.WITHDRAWAL_COLUMN_NAME)
            elif 'Debit' in table_headers:
                self.rename_column(df, 'Debit', self.WITHDRAWAL_COLUMN_NAME)
            elif 'Pay Out' in table_headers:
                self.rename_column(df, 'Pay Out', self.WITHDRAWAL_COLUMN_NAME)
            elif 'Withdrawal' in table_headers:
                self.rename_column(df, 'Withdrawal', self.WITHDRAWAL_COLUMN_NAME)

            if 'Credits' in table_headers:
                self.rename_column(df, 'Credits', self.DEPOSIT_COLUMN_NAME)
            elif 'Credit' in table_headers:
                self.rename_column(df, 'Credit', self.DEPOSIT_COLUMN_NAME)
            elif 'Deposit' in table_headers:
                self.rename_column(df, 'Deposit', self.DEPOSIT_COLUMN_NAME)
            elif 'Lodgements' in table_headers:
                self.rename_column(df, 'Lodgements', self.DEPOSIT_COLUMN_NAME)
            elif 'Pay In' in table_headers:
                self.rename_column(df, 'Pay In', self.DEPOSIT_COLUMN_NAME)

            ## rename date columns
            if 'Date' in table_headers:
                self.rename_column(df, 'Date', self.TRANSACTION_DATE_NAME)
            elif 'Trans Date' in table_headers:
                self.rename_column(df, 'Trans Date', self.TRANSACTION_DATE_NAME)
            elif 'Date Posted' in table_headers:
                self.rename_column(df, 'Date Posted', self.TRANSACTION_DATE_NAME)
            elif 'TransDate' in table_headers:
                self.rename_column(df, 'TransDate', self.TRANSACTION_DATE_NAME)
            elif 'Txn Date' in table_headers:
                self.rename_column(df, 'Txn Date', self.TRANSACTION_DATE_NAME)

            if 'ValueDate' in table_headers:
                self.rename_column(df, 'ValueDate', self.VALUE_DATE_COLUMN_NAME)
            elif 'Val Date' in table_headers:
                self.rename_column(df, 'Val Date', self.VALUE_DATE_COLUMN_NAME)

            if 'Remarks' in table_headers:
                self.rename_column(df, 'Remarks', self.DESCRIPTION_COLUMN_NAME)
            elif 'Narration' in table_headers:
                self.rename_column(df, 'Narration', self.DESCRIPTION_COLUMN_NAME)
            elif 'Transaction Details' in table_headers:
                self.rename_column(df, 'Transaction Details', self.DESCRIPTION_COLUMN_NAME)
            elif 'Details' in table_headers:
                self.rename_column(df, 'Details', self.DESCRIPTION_COLUMN_NAME)

            df['Transaction Date'] = df['Transaction Date'].apply(self.try_multiple_date_formats)
            df['Value Date'] = df['Value Date'].apply(self.try_multiple_date_formats)
            # df['Value Date'] = df['Value Date'].apply(bank_statement.clean_date_v2, format='%d-%b-%Y')
            # df['Withdrawals'] = self.clean_money('Withdrawals', df)
            df['Withdrawals'] = df['Withdrawals'].apply(self.clean_money_v2).apply(pd.to_numeric, errors='coerce')
            # df['Balance'] = self.clean_money('Balance', df)
            df['Balance'] = df['Balance'].apply(self.clean_money_v2).apply(pd.to_numeric, errors='coerce')
            # df['Deposits'] = self.clean_money('Deposits', df)
            df['Deposits'] = df['Deposits'].apply(self.clean_money_v2).apply(pd.to_numeric, errors='coerce')
            df['Description'] = self.clean_column(df, 'Description')

            # Replace NaT values with a custom string, e.g., 'Not Available'

            df['Value Date'] = df['Value Date'].fillna('Invalid Date')
            df['Transaction Date'] = df['Transaction Date'].fillna('Invalid Date')
            df['Withdrawals'] = df['Withdrawals'].fillna(0)
            df['Deposits'] = df['Deposits'].fillna(0)
            df['Balance'] = df['Balance'].fillna(0)

            return df

    @abstractmethod
    def predict_salary_income(self, dataframe, table_headers):
        pass

    def is_unique_amount_in_month_year(self, row, bank_df) -> bool:
        amount = row['Deposits']
        month_year = (row['Transaction Date'].year, row['Transaction Date'].month)
        return len(bank_df[(bank_df['Deposits'] == amount) & (bank_df['Transaction Date'].dt.year == month_year[0]) & (
                bank_df['Transaction Date'].dt.month == month_year[1])]) == 1

    @abstractmethod
    def result(self):
        pass

    def export_to_excel(self, dataframe, name, start_date, end_date, is_salary=False) -> str:
        if dataframe is None:
            return False
        try:
            # Specify the file name for the Excel file
            file_name = name.replace(' ', '_') + '_From_' + start_date + '_To_' + end_date + '.xlsx'
            directory_path = 'exports/'
            if is_salary:
                file_name = "Salary_Predictions_FOR_" + file_name
                directory_path = directory_path + "salary_predictions/"
            else:
                directory_path = directory_path + "general/"
            # Specify the directory path where you want to save the Excel file

            # if path doesn’t exist we create the path
            Path(directory_path).mkdir(parents=True, exist_ok=True)
            # Combine the directory path and file name to create the full file path
            full_file_path = directory_path + file_name
            dataframe.to_excel(full_file_path, index=False)
            return full_file_path
        except Exception as e:
            print(e)
            return False

    def get_average_monthly_balance(self, dataframe) -> float:
        if dataframe is None:
            return 0.00
        dataframe['Transaction date'] = pd.to_datetime(dataframe['Transaction Date'])
        dataframe['Balance'] = pd.to_numeric(dataframe['Balance'])
        dataframe['YearMonth'] = dataframe['Transaction Date'].dt.to_period('M')
        monthly_average_balance = dataframe.groupby('YearMonth')['Balance'].mean()
        return monthly_average_balance.mean()

    def get_predicted_salary_average(self, dataframe) -> float:
        if dataframe is None:
            return 0.00
        dataframe['Transaction date'] = pd.to_datetime(dataframe['Transaction Date'])
        dataframe['Deposits'] = pd.to_numeric(dataframe['Deposits'])
        dataframe['YearMonth'] = dataframe['Transaction Date'].dt.to_period('M')
        monthly_average_balance = dataframe.groupby('YearMonth')['Deposits'].mean()
        return monthly_average_balance.mean()

    def write_to_text_file(self, bank_name, text):
        file_directory = 'tests/pdftexts/' + bank_name + '_first_page_text.txt';
        with open(file_directory, 'w', encoding='utf-8') as f:
            f.write(text)

    def predict_repayment_capability(self, dataframe, required_payment: float):

        # Create a dictionary to store the count for each month
        result_dict = {}

        # Loop through the transactions
        for index, row in dataframe.iterrows():
            transaction_date = row['Transaction date']
            month_name = transaction_date.strftime('%B')  # Get month name

            # Update the count for the current month if balance is higher or equal to required payment
            if row['Balance'] >= required_payment:
                if month_name not in result_dict:
                    result_dict[month_name] = 1
                else:
                    result_dict[month_name] += 1

        result_array = []

        for month_name, count in result_dict.items():
            result_array.append({'month_name': month_name, 'count': count})
        print(result_array)
        return result_array



from bank_statement_reader.BaseBankStatementReport import BankStatementReport
import pandas as pd
import re


class ZenithBankStatement(BankStatementReport):

    def __init__(self, pdf_directory, min_salary, max_salary):
        if pdf_directory is None or pdf_directory == '':
            pdf_directory = "pdfs/zenith.pdf"
        super().__init__(password='', pdf_directory=pdf_directory, min_salary=min_salary, max_salary=max_salary, bank_name='zenith')

    def predict_salary_income(self, dataframe, table_headers):
        # Filter the DataFrame to get rows with values within the specified range
        filtered_df = dataframe[(dataframe['Deposits'] >= self.min_salary) & (dataframe['Deposits'] <= self.max_salary)]

        potential_salary = []
        for index, row in filtered_df.iterrows():
            unique = self.is_unique_amount_in_month_year(row, filtered_df)
            if not unique:
                continue
            potential_salary.append([
                row['Transaction Date'],
                row['Value Date'],
                row['Description'],
                row['Withdrawals'],
                row['Deposits'],
                row['Balance'],
            ])
        salary_df = self.format_dataframe_columns(table_headers, potential_salary)
        return salary_df

    def result(self):
        reader, status, message = self.get_pdf_reader()
        print(message)
        if status == 0:
            raise Exception("Reading of file failed")
        text = self.get_pdf_page_text(reader)
        cleaned_text = self.clean_text(text)

        statement_period_extracted = self.get_statement_period(cleaned_text)
        account_name_extracted = self.get_account_name(cleaned_text)
        account_number_extracted = self.get_account_number(cleaned_text)
        total_withdrawals_extracted = self.get_total_withdrawal(cleaned_text)
        total_deposit_extracted = self.get_total_deposit(cleaned_text)
        opening_balance_extracted = self.get_opening_balance(cleaned_text)
        closing_balance_extracted = self.get_closing_balance(cleaned_text)

        table_headers = self.get_transactions_table_headers(reader)

        num_pages = len(reader.pages)
        trans_rows = []
        for page_num in range(num_pages):
            try:
                new_rows = self.get_transactions_table_rows(reader, page_num)
                trans_rows.extend(new_rows)
            except Exception as e:
                print(page_num)
                print("from result", e)
        if opening_balance_extracted is None:
            opening_balance_extracted = trans_rows[0][5]

        if closing_balance_extracted is None:
            closing_balance_extracted = trans_rows[len(trans_rows) - 1][5]
        formatted_df = self.format_dataframe_columns(table_headers, table_rows=trans_rows)
        average_monthly_balance = self.get_average_monthly_balance(formatted_df.copy())
        return {
            'dataframe': formatted_df,
            'period': statement_period_extracted,
            "account_name": account_name_extracted,
            "account_number": account_number_extracted,
            "total_turn_over_credit": total_deposit_extracted,
            "total_turn_over_debits": total_withdrawals_extracted,
            "opening_balance": opening_balance_extracted,
            "closing_balance": closing_balance_extracted,
            "average_monthly_balance": average_monthly_balance
        }

    def get_account_number(self, text):
        pattern = r'Account\s+(?:No(?:\.|:)?|Number(?:\.|:)?)\s+SA\s+(\d{10,12})'
        match = re.search(pattern, text, re.IGNORECASE)
        if match:
            account_number = match.group(1)
            return account_number
        else:
            return None

    def get_account_name(self, text):

        pattern = r"(?i)(\w+)\s+(\w+)\s+(\w+)\s+Account\s+Number:"

        match = re.search(pattern, text, re.IGNORECASE)
        full_name = ''
        if match:
            first_name = match.group(1)
            middle_name = match.group(2)
            last_name = match.group(3)

            if first_name is not None:
                full_name = first_name

            if middle_name is not None:
                full_name = full_name + ' ' + middle_name

            if last_name is not None:
                full_name = full_name + ' ' + last_name
            return full_name
        else:
            return None

    def get_transactions_table_headers(self, reader):
        return [
            'Date Posted',
            'Value Date',
            'Description',
            'Debit',
            'Credit',
            'Balance'
        ]

    def add_pipe_to_money(self, text):
        # Define a regular expression to find strings that look like money (e.g., 76.29, 3,176.29, 76.29)
        # Define a regular expression to find occurrences of the pattern "date date amount amount"
        date_pattern = r'\d{2}/\d{2}/\d{4}'
        money_pattern = r'\d{1,3}(?:,\d{3})*(?:\.\d{2})?'

        full_pattern = rf'{date_pattern}\s{date_pattern}\s{money_pattern}\s{money_pattern}'

        # Use regex to find all occurrences of the money pattern in the text
        matches = re.findall(full_pattern, text)

        # Iterate through the matches and add a pipe at the end of each match
        for match in matches:
            text = text.replace(match, match + ' |')

        return text

    def insert_pipe_after_opening_balance(self, input_string):
        # Define the regular expression pattern for money
        money_pattern = r'\d{1,3}(?:,\d{3})*(?:\.\d{2})?'

        # Define the regular expression pattern for "Opening Balance" followed by money
        opening_balance_pattern = r'OPENING BALANCE\s+(' + money_pattern + ')'

        # Search for the pattern in the input string
        match = re.search(opening_balance_pattern, input_string)

        # If the pattern is found, insert a pipe "|" at the front of the money pattern
        if match:
            modified_string = input_string[:match.start(1)] + match.group(1) + ' | ' + input_string[match.end(1):]
            return modified_string
        else:
            return input_string

    def find_entries_after_header(self, data_string, header):
        data_lines = data_string.split('\n')
        header_index = None

        for i, line in enumerate(data_lines):
            if header in line:
                header_index = i
                break

        if header_index is not None:
            return '\n'.join(data_lines[header_index + 1:])
        else:
            return ""

    def extract_patterns(self, input_string):
        try:
            # Define the regular expression patterns
            date_pattern = r'\d{2}/\d{2}/\d{4}'
            money_pattern = r'[\d,]+\.\d{2}'

            # Find the first and second date strings
            dates = re.findall(date_pattern, input_string)
            first_date_str, second_date_str = dates[0], dates[1]

            # Find the money pattern(s)
            money_patterns = re.findall(money_pattern, input_string)

            # Remove the money patterns and dates from the input string
            remaining_string = re.sub(date_pattern, '', input_string)
            remaining_string = re.sub(money_pattern, '', remaining_string).strip()

            return [first_date_str, second_date_str, money_patterns, remaining_string]
        except Exception as e:
            print("Extract Pattern method: ", e)

    def get_transactions_table_rows(self, reader, page):
        header_to_find = "DATE POSTED VALUE DATE DESCRIPTION DEBIT CREDIT BALANCE"
        text = reader.pages[page].extract_text()
        date_pattern = r'\d{2}/\d{2}/\d{4}'
        money_pattern = r'[\d,]+\.\d{2}'
        rows_without_header = []
        split_balance_opening_balance = 0
        if page == 0:
            extracted_text = self.find_entries_after_header(text, header_to_find)
            extracted_text = self.insert_pipe_after_opening_balance(extracted_text)
        else:
            extracted_text = self.find_entries_after_header(text, header_to_find)

        extracted_text = self.add_pipe_to_money(extracted_text)
        # print(extracted_text)
        split_extracted_text = extracted_text.split('|')
        # something = split_extracted_text[len(split_extracted_text) - 3].replace('\n', ' ')
        for item in split_extracted_text:
            if not (item.find('OPENING BALANCE') != -1):
                cleaned_item = item.replace('\n', '|')

                list_of_cleaned_item = cleaned_item.split(' ')
                list_of_cleaned_item = self.remove_empty_and_none(list_of_cleaned_item)

                length_of_list_of_cleaned_item = len(list_of_cleaned_item)
                balance = list_of_cleaned_item[length_of_list_of_cleaned_item - 1]
                credit_or_debit = list_of_cleaned_item[length_of_list_of_cleaned_item - 2]
                value_date = list_of_cleaned_item[length_of_list_of_cleaned_item - 3]
                date_posted = list_of_cleaned_item[length_of_list_of_cleaned_item - 4]

                description = ''.join(list_of_cleaned_item[0: len(list_of_cleaned_item) - 4])
                alternative_item = self.extract_patterns(item)
                alternative_credit_or_debit = ''
                alternative_item_balance = ''
                if alternative_item is not None:
                    money_patterns = alternative_item[2]
                    if len(money_patterns) == 2:
                        # print(alternative_credit_or_debit)
                        alternative_item_balance = money_patterns[1]
                        alternative_credit_or_debit = money_patterns[0]
                    alternative_value_date = alternative_item[1]
                    alternative_date_posted = alternative_item[0]
                    # print(alternative_item)
                    alternative_item_description = alternative_item[3]

                if re.match(date_pattern, date_posted) is None and re.match(date_pattern, alternative_date_posted):
                    date_posted = alternative_date_posted

                if re.match(date_pattern, value_date) is None and re.match(date_pattern, alternative_value_date):
                    value_date = alternative_value_date
                    # description = alternative_item_description

                if re.match(money_pattern, balance) is None and re.match(money_pattern, alternative_item_balance):
                    balance = alternative_item_balance

                if re.match(money_pattern, credit_or_debit) is None and alternative_credit_or_debit is not None:
                    if re.match(money_pattern, alternative_credit_or_debit):
                        credit_or_debit = alternative_credit_or_debit
                date_posted_is_valid = re.match(date_pattern, date_posted)
                value_date_is_valid = re.match(date_pattern, value_date)
                credit_or_debit_is_valid = re.match(money_pattern, balance)
                if date_posted_is_valid and value_date_is_valid and credit_or_debit_is_valid:
                    row = [date_posted, value_date, description, credit_or_debit, credit_or_debit, balance]
                    rows_without_header.append(row)
                else:
                    continue
            else:
                split_balance_opening_balance = item.split(' ')[2]
                row = [None, None, 'Opening Balance', 0.00, 0.00, split_balance_opening_balance]
                # rows_without_header.append(row)

        for index, row in enumerate(rows_without_header):
            # row[5] is balance
            # row[4] is credit|deposit
            # row[3] is balance
            current_row = rows_without_header[index]
            previous_row = rows_without_header[index - 1]
            current_row_balance = current_row[5]
            previous_row_balance = previous_row[5]

            # if index != 0:
            current_row_balance = float(current_row_balance.replace(',', ''))
            previous_row_balance = float(previous_row_balance.replace(',', ''))
            if current_row_balance > previous_row_balance:
                row[3] = 0.00
            else:
                row[4] = 0.00

        # else:
        #     if float(split_balance_opening_balance) > 0:
        #         row[3] = 0.00
        #         row[4] = 0.00
        # print(rows_without_header)
        # exit()
        return rows_without_header

    def remove_empty_and_none(self, input_list):
        return [item for item in input_list if item is not None and item != ""]

    def get_transactions_table_header_mapping(self):
        return {
            'date_posted': 'Date Posted',
            'value_date': 'Value Date',
            'description': 'Description',
            'debit': 'Debit',
            'credit': 'Credit',
            'balance': 'Balance'
        }

# if __name__ == "__main__":
#     print("Called")
#     pdf_path = "../pdfs/zenith.pdf"
#
#     bank_statement = ZenithBankStatement(pdf_path)
#
#     print(bank_statement.result())
